import unittest
from .sarcasm import Sarcasm

from plover.formatting import _Action


class TestSarcasm(unittest.TestCase):

    def test_input_case(self):
        s = Sarcasm()
        self.assertEqual(s.format("TEST"), "tESt")

    def test_vowels_alternate(self):
        s = Sarcasm()
        self.assertEqual(s.format("eeeeeeee"), "eEeEeEeE")

    def test_consonants_alternate(self):
        s = Sarcasm()
        self.assertEqual(s.format("bbbbbbbb"), "bBbBbBbB")

    def test_whitespace(self):
        s = Sarcasm()
        self.assertEqual(s.format(" test"), " tESt")

    def test_punctuation(self):
        s = Sarcasm()
        self.assertEqual(s.format("(test)"), "(tESt)")

    def test_oh_really(self):
        s = Sarcasm()
        self.assertEqual(s.format("oh really?"), "oH rEaLlY?")

    def test_nice_jacket(self):
        s = Sarcasm()
        self.assertEqual(s.format("nice jacket"), "nICe jACkET")

    def test_look_great(self):
        s = Sarcasm()
        self.assertEqual(s.format("you look great"), "yOu LoOk GrEaT")

    def test_borrow_pencil(self):
        s = Sarcasm()
        self.assertEqual(s.format("can I borrow a pencil?"),
                         "cAN i bORrOW a pENcIL?")

    def test_replace_action(self):
        s = Sarcasm()
        a = [
            _Action(text="previous text"),
            _Action(prev_replace="previous text", text="new text")
        ]
        s.process_actions(a)
        self.assertTrue(a[0].text.endswith(a[1].prev_replace))



from .formatterbase import FormatterBase


class Substitute(FormatterBase):

    def __init__(self, character_map, joiner=''):
        self._joiner = joiner
        self._character_map = character_map

    def swap(self, c: str) -> str:
        if c in self._character_map:
            return self._character_map[c]
        return c

    def format(self, str: str) -> str:
        if str is None:
            return None
        return self._joiner.join(self.swap(c) for c in str)



from typing import List
from plover.formatting import _Action


class FormatterBase():

    # default no-state reset
    def reset_state(self):
        pass

    # decent default implementation of processing actions
    # stateful formatters may want to do something different
    # or may not
    def process_actions(self, new: List[_Action]):
        for a in new:
            if a.combo == 'Return':
                self.reset_state()

            if a.text:
                a.text = (self.format(a.text))
            if a.word:
                a.word = (self.format(a.word))
            if a.prev_replace:
                a.prev_replace = self.format(a.prev_replace)



from typing import List
from threading import Thread

from plover.engine import StenoEngine
from plover import log
from plover.registry import registry
from plover.formatting import _Context, _Action

from .common import TRANSFORMERS


class ExtensionPlugin(Thread):

    def __init__(self, engine: StenoEngine) -> None:
        super().__init__()

        log.info("FANCY_INIT")
        registry.register_plugin('meta', 'fancytext_set', self.fancy_set)

        self.formatter = None
        self.engine = engine
        self.transformers = TRANSFORMERS

    def fancy_set(self, ctx: _Context, cmdline) -> _Action:
        args = cmdline.split(':')
        mode = args.pop(0)
        if mode in self.transformers:
            self.formatter = self.transformers[mode](*args)
        else:
            self.formatter = None

        return ctx.new_action()

    def start(self) -> None:
        log.info("FANCY_START")
        self.engine.hook_connect("translated", self.translated)
        super().start()

    def stop(self) -> None:
        self.engine.hook_disconnect("translated", self.translated)

    def translated(self, old: List[_Action], new: List[_Action]) -> None:
        if self.formatter:
            self.formatter.process_actions(new)



from typing import List
from plover.formatting import _Action

import random

from .character_helpers import COMBINING_MARKS
from .formatterbase import FormatterBase


class Zalgo(FormatterBase):

    def __init__(self, minimum, maximum):
        self.minimum = minimum
        self.maximum = maximum
        assert self.minimum <= self.maximum
        self.reset_state()

    def reset_state(self):
        self.translations = {}

    def add_combining_marks(self, c: str) -> str:
        # In order for suffix folding to work, going to need some determinism
        # otherwise Assert Failed! in formatting.py
        if c in self.translations:
            return self.translations[c]

        for m in random.sample(
                COMBINING_MARKS,
                random.randrange(self.minimum, self.maximum)):
            o = c
            o += m
            self.translations[c] = o
        return o

    def format(self, str) -> str:
        if str:
            return ''.join(self.add_combining_marks(s) for s in str)
        return None

    def process_actions(self, new: List[_Action]):
        for a in new:
            if a.text:
                a.text = (self.format(a.text))
            if a.word:
                a.word = (self.format(a.word))
            if a.prev_replace:
                a.prev_replace = self.format(a.prev_replace)
        # can reset every set of actions
        # the memoization is so that prev_replace won't throw errors
        # if prev_replace isn't at the end of the last action's .text
        self.reset_state()



from typing import List
from plover.formatting import _Action

from .formatterbase import FormatterBase

from pyfiglet import Figlet


class Figletise(FormatterBase):

    def __init__(self, font):
        self.fig = Figlet(font)

    def format(self, str) -> str:
        if str:
            return self.fig.renderText(str)
        return None

    def process_actions(self, new: List[_Action]):
        raise Exception(
            "Mode not supported, use fancy_text_retro:<count>:figlet:<font>"
        )



import unittest

from .zalgo import Zalgo


class TestZalgo(unittest.TestCase):

    def test_original_included(self):
        z = Zalgo(5, 10)
        zalgo = ["z", "a", "l", "g", "o"]
        output = z.format("".join(zalgo))

        for c in zalgo:
            self.assertIn(c, output)



from unittest import TestCase
from .retro import fancytext_retro

from plover.formatting import _Action


class FakeContext():
    def __init__(self, actions):
        self.actions = actions

    def copy_last_action(self):
        return self.actions[-1]

    def last_words(self, count):
        return [a.word for a in self.actions[-count:]]


class TestRetro(TestCase):

    def test_retro_sarcasm(self):
        action = _Action(word="fancy", text="fancy")
        retroed = fancytext_retro(FakeContext([action]), "1:sarcasm")

        self.assertEqual(retroed.prev_replace, "fancy")
        self.assertEqual(retroed.text, "fANcY")
        self.assertIsNone(retroed.word)
        self.assertTrue(retroed.prev_attach)



from typing import List
from plover.formatting import _Action

from .character_helpers import CONSONANT_RE, VOWEL_RE
from .formatterbase import FormatterBase


class Sarcasm(FormatterBase):

    # the first letter must be a lower
    # Every second consonant must be upper
    # if vowel after a lower it must be upper
    # if vowel comes after an upper it must be lower

    def __init__(self):
        self.reset_state()

    def reset_state(self):
        self.first_letter = True
        self.capped_last_consonant = False
        self.capped_last = False

    def case_letter(self, c: str) -> str:
        # first letter is a special case
        if self.first_letter:
            if VOWEL_RE.match(c):
                self.first_letter = False

            if CONSONANT_RE.match(c):
                self.first_letter = False
            return c.lower()

        # consonants alternate
        if CONSONANT_RE.match(c):
            if self.capped_last_consonant:
                self.capped_last_consonant = False
                self.capped_last = False
                return c.lower()

            self.capped_last = True
            self.capped_last_consonant = True
            return c.upper()

        if VOWEL_RE.match(c):
            if self.capped_last:
                self.capped_last = False
                return c.lower()
            else:
                self.capped_last = True
                return c.upper()

        return c

    def format(self, str) -> str:
        return ''.join(self.case_letter(s) for s in str)

    def process_actions(self, new: List[_Action]):
        for i, a in enumerate(new):
            if a.combo == 'Return':
                self.reset_state()

            if a.text:
                a.text = (self.format(a.text))
            if a.prev_replace:
                a.prev_replace = new[i-1].text[-len(a.prev_replace):]



import re
from .formatterbase import FormatterBase

UWU_SUBSTITUTIONS = {
    "l": "w",
    "r": "w",
    "v": "w",
}

INTENSE_UWU_SUBSTITUTIONS = {
    "u": "uwu",
    ".": " :3"
}


class Uwu(FormatterBase):
    def __init__(self, intense: bool):
        self.intense = intense

    def uwuify_char(self, char: str) -> str:
        result = self.substitute_with_map(UWU_SUBSTITUTIONS, char)
        if self.intense:
            result = self.substitute_with_map(INTENSE_UWU_SUBSTITUTIONS,
                                              result)
        return result

    def substitute_with_map(self, substitutions, char: str) -> str:
        result = substitutions.get(char.lower())
        if result is None:
            return char
        elif char.isupper():
            return result.capitalize()

        return result

    def format(self, string: str) -> str:
        string = re.sub("([nN])([aeiouAEIOU])", r"\1y\2", string)
        string = string \
            .replace("tute", "tewt") \
            .replace("Hello", "Hewwoooo <3") \
            .replace("yes", "yaww") \
            .replace("no", "nouu")
        return ''.join([self.uwuify_char(char) for char in string])



from typing import List
import random

from .formatterbase import FormatterBase
from .character_helpers import LETTERS

from plover.formatting import _Action


class CryTyping(FormatterBase):

    # swapping two letters
    # adding a random letter
    # doubling a letter
    # adding random punctuation
    # deleting a letter
    # swapping the case of a letter (e.g. a ⇔ A)

    def __init__(self):
        self.reset_state()

    def reset_state(self):
        self.letter_swap = None

    def letter_operations(self, c: str) -> str:
        o = c

        if self.letter_swap:
            o += self.letter_swap
            self.reset_state()
            return o

        if random.randint(1, 100) > 98:
            # delete :D
            return ''

        upper_case_chance = random.randint(1, 100)

        if upper_case_chance > 97:
            o = o.upper()

        swap_chance = random.randint(1, 100)

        if swap_chance > 96:
            self.letter_swap = c
            return ''

        dupe_chance = random.randint(1, 100)
        if dupe_chance > 90:
            o += c

        punc_chance = random.randint(1, 100)
        punc_amount = random.randint(2, 4)

        if punc_chance > 95:
            p = random.choice([',', '.', ';'])
            for x in range(1, punc_amount):
                o += p

        add_random_chance = random.randint(1, 100)

        if add_random_chance > 98:
            o += random.choice(LETTERS)
        return o

    def format(self, str) -> str:
        if str:
            letters = (self.letter_operations(c)
                       for c in str)
            self.reset_state()  # it's ok if we lose a letter
            return ''.join(letters)
        return None

    def process_actions(self, new: List[_Action]):
        for a in new:
            if a.text:
                a.text = (self.format(a.text))
            if a.word:
                a.word = (self.format(a.word))
            if a.prev_replace:
                # if this doesn't match the end of the previous action
                # then plover will throw assertion errors
                # galaxy brain: just get rid of it
                a.prev_replace = None



from .zalgo import Zalgo
from .uwu import Uwu
from .crytyping import CryTyping
from .sarcasm import Sarcasm
from .substitute import Substitute
from .figlet import Figletise
from .character_helpers import \
    BUBBLE_MAP, MEDIEVAL_MAP, UPSIDE_DOWN_MAP,\
    FULLWIDTH_MAP, MORSE_MAP, SMALLCAPS_MAP,\
    SCRIPT_MAP, BLACKBOARD_BOLD_MAP, MONOSPACE_MAP

TRANSFORMERS = {
    'bubble': lambda: Substitute(BUBBLE_MAP),
    'morse': lambda: Substitute(MORSE_MAP, ' '),
    'crytyping': lambda: CryTyping(),
    'medieval': lambda: Substitute(MEDIEVAL_MAP),
    'fullwidth': lambda: Substitute(FULLWIDTH_MAP),
    'uwu': lambda: Uwu(intense=False),
    'UwU': lambda: Uwu(intense=True),
    'sarcasm': lambda: Sarcasm(),
    'figlet': lambda font='slant': Figletise(font),
    'upsidedown': lambda: Substitute(UPSIDE_DOWN_MAP),
    'smallcaps': lambda: Substitute(SMALLCAPS_MAP),
    'script': lambda: Substitute(SCRIPT_MAP),
    'blackboardbold': lambda: Substitute(BLACKBOARD_BOLD_MAP),
    'monospace': lambda: Substitute(MONOSPACE_MAP),
    'zalgo': lambda minimum=1, maximum=3: Zalgo(int(minimum),
                                                int(maximum))
}



from plover.formatting import _Context, _Action

from .common import TRANSFORMERS


def fancytext_retro(ctx: _Context, cmdline: str) -> _Action:
    action = ctx.copy_last_action()
    args = cmdline.split(':')

    # should be <numwords>:mode:<any extra mode args>
    count = int(args.pop(0))
    mode = args.pop(0)

    transformer = TRANSFORMERS[mode](*args)

    last_words = ''.join(ctx.last_words(count=count))

    action.word = None
    action.text = transformer.format(last_words)

    action.prev_replace = last_words
    action.prev_attach = True

    return action



import unittest
import inspect

from plover.formatting import _Action

from .extensionplugin import ExtensionPlugin


class FakeContext():

    def new_action(self):
        pass


class FakeEngine():

    def __init__(self):
        self.method_calls = {}

    def hook_connect(self, hook: str, method):
        self.method_calls[hook] = method

    def hook_disconnect(self, hook: str, method):
        self.method_calls[hook] = None


class TestExtensionPlugin(unittest.TestCase):

    def test_connect_hook(self):
        e = FakeEngine()
        f = ExtensionPlugin(e)
        f.start()
        self.assertIsNotNone(e.method_calls["translated"])

    def test_hook_signature(self):
        e = FakeEngine()
        f = ExtensionPlugin(e)
        f.start()
        hooked_method = e.method_calls["translated"]
        signature = inspect.signature(hooked_method)

        self.assertEqual(len(signature.parameters), 2)
        self.assertIsNotNone(signature.parameters["old"])
        self.assertIsNotNone(signature.parameters["new"])

    def test_disconnect_hook(self):
        e = FakeEngine()
        f = ExtensionPlugin(e)
        f.start()
        f.stop()
        self.assertIsNone(e.method_calls["translated"])

    def test_translate_does_nothing_if_fancy_not_set(self):
        p = self._set_up_plugin()
        old = []
        old.append(self._get_action_return("things"))
        new = []
        new.append(self._get_action_return("stuff"))

        p.translated(old, new)

        self.assertEqual(len(old), 1)
        self.assertEqual(old[0], self._get_action_return("things"))
        self.assertEqual(len(new), 1)
        self.assertEqual(new[0], self._get_action_return("stuff"))

    def test_plugged_bubble(self):
        p = self._set_up_plugin()
        p.fancy_set(FakeContext(), "bubble")
        new = self._get_action_return("text")
        p.translated([], new)

        self.assertEqual(new[0].text, "ⓣⓔⓧⓣ")

        new = self._get_orthography_thing()
        p.translated([], new)
        self.assertTrue(new[0].text.endswith(new[1].prev_replace))

    def test_plugged_crytyping(self):
        p = self._set_up_plugin()
        p.fancy_set(FakeContext(), "crytyping")
        # this test .. uh makes sure it doesn't blow up?
        p.translated([], self._get_action_return("text"))
        new = self._get_orthography_thing()
        p.translated([], new)

        # crytyping is too random for prev_replace to work
        self.assertIsNone(new[1].prev_replace)

    def test_plugged_medieval(self):
        p = self._set_up_plugin()
        p.fancy_set(FakeContext(), "medieval")
        new = self._get_action_return("F")
        p.translated([], new)

        self.assertEqual(new[0].text, "𝕱")

    def test_plugged_fullwidth(self):
        p = self._set_up_plugin()
        p.fancy_set(FakeContext(), "fullwidth")
        new = self._get_action_return("F")
        p.translated([], new)

        self.assertEqual(new[0].text, "\uFF26")

    def test_plugged_uwu(self):
        p = self._set_up_plugin()
        p.fancy_set(FakeContext(), "uwu")
        new = self._get_action_return("McNugget")
        p.translated([], new)

        self.assertEqual(new[0].text, "McNyugget")

    def test_plugged_uwu_intense(self):
        p = self._set_up_plugin()
        p.fancy_set(FakeContext(), "UwU")
        new = self._get_action_return(".")
        p.translated([], new)

        self.assertEqual(new[0].text, " :3")

    def test_plugged_sarcasm(self):
        p = self._set_up_plugin()
        p.fancy_set(FakeContext(), "sarcasm")
        new = self._get_action_return("testing")
        p.translated([], new)

        self.assertEqual(new[0].text, "tEStINg")

    def test_plugged_upsidedown(self):
        p = self._set_up_plugin()
        p.fancy_set(FakeContext(), "upsidedown")
        new = self._get_action_return("test")
        p.translated([], new)

        self.assertEqual(new[0].text, "\u0287\u01DD\u0073\u0287")

    def test_plugged_zalgo(self):
        p = self._set_up_plugin()
        p.fancy_set(FakeContext(), "zalgo")
        new = self._get_orthography_thing()
        p.translated([], new)

        self.assertTrue(new[0].text.endswith(new[1].prev_replace))

    def test_plugged_morse(self):
        p = self._set_up_plugin()
        p.fancy_set(FakeContext(), "morse")
        new = self._get_action_return("test")
        p.translated([], new)

        self.assertEqual(new[0].text, "- . ... -")

    def _set_up_plugin(self) -> ExtensionPlugin:
        e = FakeEngine()
        p = ExtensionPlugin(e)
        p.start()
        return p

    def _get_action_return(self, str):
        return [
            _Action(text=str, word=str),
            _Action(combo="Return")
        ]

    def _get_orthography_thing(self):
        return [
            _Action(text="the", word="the"),
            _Action(text="ing", word="thing", prev_replace="e")
        ]
